;; Generated package description from activities.el  -*- no-byte-compile: t -*-
(define-package "activities" "0.4.0.20240201.202903" "Save/restore sets of windows, tabs/frames, and their buffers" '((emacs "29.1") (persist "0.6")) :commit "3e1ba4e12531f2e01a1b29ac5d53ff543a210d03" :authors '(("Adam Porter" . "adam@alphapapa.net")) :maintainer '("Adam Porter" . "adam@alphapapa.net") :keywords '("convenience") :url "https://github.com/alphapapa/activities.el")
